-- tag::code_create_ch13_staging[] --
CREATE SCHEMA ch13_staging;
CREATE SCHEMA ch13;
-- end::code_create_ch13_staging[] --

-- tag::code_create_topology_4326[] --
SELECT CreateTopology('ch13a_topology',4326); 
-- end::code_create_topology_4326[] --

-- tag::code_load_colorado[] -- 
SELECT TopoGeo_AddLineString(
    'ch13a_topology', 
	ST_GeomFromText(
		'LINESTRING(
			-109.05304 39.195013,
			-109.05304 41.000889,
			-104.897461 40.996484
		)',
		4326
    )
);

SELECT TopoGeo_AddLineString(
	'ch13a_topology',
	ST_GeomFromText(
		'LINESTRING(
		    -104.897461 40.996484,
		    -102.051744 40.996484, 
		    -102.051744 40.003029
		)',
		4326
    )
);

SELECT TopoGeo_AddLineString(
	'ch13a_topology',
	ST_GeomFromText(
		'LINESTRING(
		    -102.051744 40.003029,
		    -102.04874 36.992682,
		    -104.48204 36.992682
		)',
		4326
    )
);
     
SELECT TopoGeo_AddLineString(
	'ch13a_topology',
	ST_GeomFromText(
		'LINESTRING(
		    -104.48204 36.992682,
		    -109.045226 36.999077, 
		    -109.05304 39.195013
		)',
		4326
    )
 );
-- end::code_load_colorado[] --	
	
-- tag::code_load_colorado_highway_I-70[] --
SELECT TopoGeo_AddLineString(
	'ch13a_topology',
	ST_GeomFromText(
		'LINESTRING(
			-109.05304 39.195013, 
			-108.555908 39.108751, 
			-105.021057 39.717751, 
			-102.051744 40.003029
		)',
		4326
    )
 );
-- end::code_load_colorado_highway_I-70[] --

-- tag::code_load_colorado_highway_I-25[] --
SELECT TopoGeo_AddLineString(
	'ch13a_topology',
	ST_GeomFromText(
		'LINESTRING(
			-104.897461 40.996484,
			-105.021057 39.717751,
			-104.798584 38.814031,
			-104.48204 36.992682
		)',
		4326
    )
);
-- end::code_load_colorado_highway_I-25[] --

-- tag::code_faces_colorado[] --
SELECT face_id, mbr,
ST_GetFaceGeometry('ch13a_topology',face_id) AS geom -- <1>
FROM ch13a_topology.face 
WHERE face_id > 0; -- <2>
-- end::code_faces_colorado[] --
-- #1 output face_id and face geometry for each face
-- #2 Exclude universal face

-- tag::query_list_topologies[] --
SELECT id, name, srid, precision, hasz FROM topology.topology;
-- end::query_list_topologies[] --

-- tag::code_highways_topogeom[] --
CREATE TABLE ch13.highways_topo (highway varchar(20) PRIMARY KEY); 
SELECT AddTopoGeometryColumn(
    'ch13a_topology',
    'ch13',
    'highways_topo',
    'topo',
    'LINESTRING'
);
-- end::code_highways_topogeom[] --

-- tag::code_highways_createtopogeom[] --
INSERT INTO ch13.highways_topo (highway, topo) 
VALUES (
    'I70', 
    CreateTopoGeom( -- <1>
        'ch13a_topology',
        2, -- <2>
        1, -- <3>
        '{{5,2},{6,2}}'::topoelementarray -- <4>
    )
);
-- end::code_highways_createtopogeom[] --
  
-- tag::code_highways_totopogeom[] --
INSERT INTO ch13.highways_topo (highway, topo) 
SELECT 
    'I25', 
	toTopoGeom( -- <1>
		ST_GeomFromText(
			'LINESTRING(
                -104.897461 40.996484,
                -105.021057 39.717751,
                -104.798584 38.814031,
                -104.48204 36.992682
            )',
			4326
		), -- <2>
	'ch13a_topology', -- <3>
	1 -- <4>
  );
-- end::code_highways_totopogeom[] --

-- tag::code_query_topology_co_topo_elements[] --
SELECT highway, (topo).*, GetTopoGeomElements(topo) As el
FROM ch13.highways_topo
ORDER BY highway;
-- end::code_query_topology_co_topo_elements[] --

-- tag::output_topology_co_topo_elements[] --
 highway | topology_id | layer_id | id | type |  el
---------+-------------+----------+----+------+-------
 I25     |           1 |        1 |  2 |    2 | {7,2}
 I25     |           1 |        1 |  2 |    2 | {8,2}
 I70     |           1 |        1 |  1 |    2 | {5,2}
 I70     |           1 |        1 |  1 |    2 | {6,2}
-- end::output_topology_co_topo_elements[] --

-- tag::code_create_ch13_topology[] --
SELECT CreateTopology('ch13_topology', 32610, 0.05);
-- end::code_create_ch13_topology[] --

-- tag::code_create_ch13_admin_boundaries_line[] --
SELECT 
    gid, 
    TopoGeo_AddLineString(
        'ch13_topology', ST_Transform(geom, 32610)
    ) As edge_id -- <1>
FROM (
    SELECT gid, (ST_Dump(geom)).geom FROM ch13_staging.cityboundary
) As f; -- <2>
-- end::code_create_ch13_admin_boundaries_line[] --

-- tag::code_count_faces_edges_nodes_relations[] --
SELECT 'faces' As type, COUNT(*) As num FROM ch13_topology.face
UNION ALL
SELECT 'edges' As type, COUNT(*) As num FROM ch13_topology.edge
UNION ALL
SELECT 'nodes' As type, COUNT(*) As num FROM ch13_topology.node
UNION ALL
SELECT 'relations' As type, COUNT(*) As num FROM ch13_topology.relation;
-- end::code_count_faces_edges_nodes_relations[] --

-- tag::code_create_topogeo_addpolygon_municipal[] --
SELECT 
    gid, 
    TopoGeo_AddPolygon(
        'ch13_topology', ST_Transform(geom, 32610), 0.05
    ) As face_id
FROM (
    SELECT 
        gid, 
        (ST_Dump(geom)).geom 
    FROM ch13_staging.neighbourhoods
) As f;
-- end::code_create_topogeo_addpolygon_municipal[] --

-- tag::code_build_topogeometry_columns_prim_child[] --
CREATE TABLE ch13.neighbourhoods (feat_name varchar(50) primary key);
SELECT AddTopoGeometryColumn(
    'ch13_topology',
    'ch13',    
    'neighbourhoods',
    'topo',
    'MULTIPOLYGON'
); -- <1>
  
CREATE TABLE ch13.cities (feat_name varchar(150) primary key);
SELECT AddTopoGeometryColumn(
    'ch13_topology', 
    'ch13',   
    'cities',
    'topo',
    'MULTIPOLYGON',
    1
);  -- <2>
-- end::code_build_topogeometry_columns_prim_child[] --

-- tag::code_totopogeom_nei[] --
INSERT INTO ch13.neighbourhoods (feat_name, topo)
SELECT  
    neighbourh,  
    toTopoGeom(
        ST_Transform(geom, 32610), 'ch13_topology', 1, 0.05
    )
FROM ch13_staging.neighbourhoods;
-- end::code_totopogeom_nei[] --

-- tag::code_createtopogeom_city[] --
INSERT INTO ch13.cities (feat_name, topo)
SELECT 
    'Victoria',    
    CreateTopoGeom(
        'ch13_topology',
        3, -- <1>
        2, -- <2>
        (
            SELECT TopoElementArray_Agg(
				ARRAY[(topo).id,(topo).layer_id]
			) -- <3>
            FROM ch13.neighbourhoods
        )
);
-- end::code_createtopogeom_city[] --

-- tag::code_city_asgeom[] --
SELECT topo::geometry FROM ch13.cities WHERE feat_name = 'Victoria';
-- end::code_city_asgeom[] --

-- tag::code_getfacegeom_city[] --
SELECT face_id, ST_GetFaceGeometry('ch13_topology', face_id)
FROM (
    SELECT (GetTopoGeomElements(topo))[1] As face_id 
    FROM ch13.cities 
    WHERE feat_name = 'Victoria'
) As x;
-- end::code_getfacegeom_city[] --

-- tag::code_count_faces_neighbourhood[] --
SELECT feat_name, COUNT(face_id) As num_faces, 
    MIN(
        ST_Area(ST_GetFaceGeometry('ch13_topology',face_id))
    )::numeric(10,2) As min_area,
    MAX(
        ST_Area(ST_GetFaceGeometry('ch13_topology',face_id))
    )::numeric(10,2) As max_area
FROM (
    SELECT feat_name, (GetTopoGeomElements(topo))[1] As face_id 
	FROM ch13.neighbourhoods
) As x
GROUP BY feat_name 
HAVING COUNT(face_id) > 1
ORDER BY COUNT(face_id) DESC;
-- end::code_count_faces_neighbourhood[] --

-- tag::code_remove_small_pockets[] --
DO 
LANGUAGE plpgsql 
$$
DECLARE r record; var_face integer;
BEGIN
    FOR r IN ( 
    SELECT DISTINCT abs(
        (ST_GetFaceEdges(
            'ch13_topology',face_id)
        ).edge
    ) As edge 
    FROM (
        SELECT feat_name, (GetTopoGeomElements(topo))[1] As face_id 
        FROM ch13.neighbourhoods
    ) As x
    WHERE ST_Area(ST_GetFaceGeometry('ch13_topology',face_id)) < 55000
     ) 
    LOOP
        BEGIN
            var_face := ST_RemEdgeNewFace('ch13_topology',r.edge);
            EXCEPTION
                WHEN OTHERS THEN
            RAISE WARNING 'Failed remove edge: %, %', r.edge, SQLERRM;
        END;
    END LOOP;
END
$$;
-- end::code_remove_small_pockets[] --

-- tag::code_determine_faces_shared[] --
SELECT (GetTopoGeomElements(topo))[1] As face_id -- <1>
FROM ch13.neighbourhoods
WHERE feat_name = 'North Park' 
INTERSECT
SELECT (GetTopoGeomElements(topo))[1] As face_id  -- <2>
FROM ch13.neighbourhoods
WHERE feat_name = 'Burnside';
-- end::code_determine_faces_shared[] --

-- tag::code_removing_faces_topogeom[] --
DELETE FROM ch13_topology.relation AS r -- <1>
WHERE EXISTS (
	SELECT topo -- <2>
	FROM ch13.neighbourhoods As n 
	WHERE 
		feat_name = 'North Park' AND 
		(topo).id = r.topogeo_id AND 
		r.element_id = 21 AND 
		r.element_type = 3
);

DELETE FROM ch13_topology.relation AS r -- <3>
WHERE EXISTS (
	SELECT topo 
	FROM ch13.neighbourhoods As n 
	WHERE 
		feat_name = 'Burnside' AND 
		(topo).id = r.topogeo_id AND 
		r.element_id = 28 AND 
		r.element_type = 3
);
-- end::code_removing_faces_topogeom[] --

-- tag::code_addtopogeomcolumn_streets[] --
CREATE TABLE ch13.streets (
	gid integer primary key, 
	feat_name varchar(50), 
	access varchar(20), 
	rd_class varchar(20), 
	max_speed numeric(10,2)
); -- <1>
  
SELECT AddTopoGeometryColumn(
	'ch13_topology',
	'ch13',
	'streets',
	'topo',
	'MULTILINESTRING'
); -- <2>
  
CREATE TABLE ch13.log_street_failures (
	gid integer primary key,
	error text
); -- <3> 
-- end::code_addtopogeomcolumn_streets[] --  

-- tag::code_totopogeom_streets[] --
CREATE OR REPLACE PROCEDURE ch13.load_streets(param_num integer) AS
$$
DECLARE r record;
BEGIN
    FOR r IN 
		SELECT * 
		FROM ch13_staging.streetcentrelines -- <1>
        ORDER BY gid 
		LIMIT param_num OFFSET (SELECT MAX(gid) from ch13.streets)
	LOOP  
		BEGIN
			INSERT INTO ch13.streets (
				gid,feat_name,access,rd_class,max_speed,topo) -- <2>
			SELECT 
				r.gid,r.streetname,r.access,r.rd_class,
				r.max_speed::numeric,
				toTopoGeom(ST_Transform(ST_Force2D(r.geom),32610),-- <3>
      'ch13_topology',3,0.05); -- <4>
			EXCEPTION WHEN OTHERS THEN
				INSERT INTO ch13.log_street_failures (gid,error) -- <5>
				VALUES (r.gid,SQLERRM);
				RAISE WARNING 
					'Loading of record % failed: %',
					r.gid,
					SQLERRM;
		END;
        COMMIT; -- <6>
    END LOOP;
END
$$
LANGUAGE plpgsql;
-- end::code_totopogeom_streets[] --


-- tag::code_geom_simplification[] --
SELECT feat_name, ST_Simplify(topo::geometry,150) As geom_simp 
FROM ch13.neighbourhoods
-- end::code_geom_simplification[] -- 

-- tag::code_topo_simplification[] --
SELECT feat_name, ST_Simplify(topo,150) As topo_simp 
FROM ch13.neighbourhoods
-- end::code_topo_simplification[] -- 

-- tag::code_validatetopology_victoria[] --
SELECT * 
FROM ValidateTopology('ch13_topology');
-- end::code_validatetopology_victoria[] --

-- tag::code_topologysummary_victoria[] --
SELECT TopologySummary('ch13_topology');
-- end::code_topologysummary_victoria[] --